/*HEADER------------------------------------------------------------------------
|
| PROGRAM:  solve_linear_equations_reg (macro)
||
| PURPOSE:  This macro solves system of linear equation without PROC IML.
                       It uses PROC REG with SSCP input dataset instead and it requieres 
                       only SAS STAT.
                       To use this approach the system of linear equations has to 
                        have "left" matix in the equation that satisfy the following conditions:

                             a) Symmetric
                             b) Positive definite

					  This can be used to solve system of equations that appear in the 
                       Newton-Raphson iteration method. In this case the "left" matrix is
                      Hessian matrix which is both symmetric and positive definite.

 Usage:

%LET tbl_name_left_in=DPLocal.&prefix.Hessian;
%LET tbl_name_right_in=DPLocal.&prefix.GRADIENT_T;
%LET tbl_name_out=Solution;
%LET tbl_name_ginv_out=Left_Matrix_Inv;  

%solve_linear_equations_reg(tbl_name_left_in=&tbl_name_left_in.
                                            ,tbl_name_right_in=&tbl_name_right_in.
										    ,tbl_name_ginv_out=&tbl_name_ginv_out.
                                            ,tbl_name_out=Solution);

%solve_linear_equations_reg(tbl_name_left_in=LDATA.Hessian_det_0
                                             ,tbl_name_right_in=LDATA.Gradient_d2
                                            , tbl_name_out=Solution2);

	First version:  01/04/2017, Yury Vilk. 
+------------------------------------------------------------------------HEADER*/

  %MACRO solve_linear_equations_reg(tbl_name_left_in=
                                                              ,tbl_name_right_in=
                                                              ,tbl_name_out=Solution
															  ,tbl_name_ginv_out=__Left_Matrix_Ginv
                                                              ,PRINT=N);

%GLOBAL DEBUG_FLAG2;
				   /* Code to suppress unnessary messages in the log except when in the debug mode.*/
 %LOCAL _DEBUG_OPTIONS_L;
 %LET _DEBUG_OPTIONS_L=%sysfunc(getoption(symbolgen)) %sysfunc(getoption(mprint)) %sysfunc(getoption(mlogic)) %sysfunc(getoption(notes)) %sysfunc(getoption(source2));

 %IF &DEBUG_FLAG2. NE  Y %THEN %DO;OPTIONS NOMLOGIC NOMPRINT NOSYMBOLGEN  NOSOURCE2; %END;


   %PUT;
   %PUT -----   Start of the macro solve_linear_equations_reg. ;
   %PUT;

%GLOBAL m_return_cd  ;
 %LET m_return_cd=0;

  %LOCAL pos_l tbl_name_left_in_root cnt_num_vars_l  nlobs_l  var_lst_l fst_var_numb_l	fst_var_nm_l  var_lst_l_2
                 pos_r  tbl_name_left_in_root cnt_num_vars_r  nlobs_r  var_rst_r fst_var_numb_r;

	  %LET pos_l=%INDEX(%STR(&tbl_name_left_in.),%STR(%());
	  %LET pos_r=%INDEX(%STR(&tbl_name_right_in.),%STR(%());

 
%IF &pos_l. >1 %THEN %LET tbl_name_left_in_root=%SUBSTR(&tbl_name_left_in.,1,%EVAL(&pos_l.-1) );
%ELSE  							  %LET tbl_name_left_in_root=&tbl_name_left_in.;

%IF &pos_r. >1 %THEN %LET tbl_name_right_in_root=%SUBSTR(&tbl_name_right_in.,1,%EVAL(&pos_r.-1) );
%ELSE  							  %LET tbl_name_right_in_root=&tbl_name_right_in.;

%IF      %LENGTH(&tbl_name_left_in.)=0  OR %sysfunc(exist(&tbl_name_left_in_root.))=0
    OR %LENGTH(&tbl_name_right_in.)=0 OR  %sysfunc(exist(&tbl_name_right_in_root.))=0
 %THEN 
			%DO; 
			%LET m_return_cd=1;
			%LET m_return_msg=One of the requiered  input tables does not exist tbl_name_left_in=&tbl_name_left_in.  tbl_name_right_in=&tbl_name_right_in. ;
  %GOTO endmac; 
   %END;

	PROC DATASETS NOLIST NOWARN;
	 DELETE _CONT_L _CONT_R __pseudo_SSCP __PE __PE_T __GINV ;
	RUN;
	QUIT;

 PROC CONTENTS DATA=&tbl_name_left_in. 
                                 OUT=_CONT_L(WHERE=(TYPE=1))  NOPRINT;
 RUN;

 PROC CONTENTS DATA=&tbl_name_right_in. 
                                 OUT=_CONT_R(WHERE=(TYPE=1))  NOPRINT;
 RUN;


 PROC SQL NOPRINT ;
		   SELECT  STRIP(PUT(COUNT(NAME),8.))
                          ,STRIP(PUT(MAX(NOBS),16.) )
                                   INTO  :cnt_num_vars_l , :nlobs_l
			 FROM  _CONT_L			    
				  ;
	  SELECT  STRIP(NAME), VARNUM  
                            INTO  : var_lst_l SEPARATED BY " " , : fst_var_numb_l
			 FROM  _CONT_L
 			  ORDER BY VARNUM
				  ;
 QUIT;

 PROC SQL NOPRINT ;
		   SELECT  STRIP(PUT(COUNT(NAME),8.))
                          ,STRIP(PUT(MAX(NOBS),16.) )
                                   INTO  :cnt_num_vars_r , :nlobs_r
			 FROM  _CONT_R			    
				  ;
	  SELECT  STRIP(NAME), VARNUM  
                            INTO  : var_lst_r SEPARATED BY " " , : fst_var_numb_r
			 FROM  _CONT_R
 			  ORDER BY VARNUM
				  ;
 QUIT;

 /*
 %PUT cnt_num_vars_l=&cnt_num_vars_l.  nlobs_l=&nlobs_l. 
             var_lst_l=&var_lst_l. fst_var_numb_l=&fst_var_numb_l.;

 %PUT cnt_num_vars_r=&cnt_num_vars_r.  nlobs_r=&nlobs_r. 
             var_lst_r=&var_lst_r. fst_var_numb_r=&fst_var_numb_r.;
  */

		%IF  &cnt_num_vars_l. ^= &nlobs_l. %THEN 
			%DO; 
   			   %LET m_return_cd=2;
			   %LET m_return_msg=Left matrix has to be symmetric. cnt_num_vars_l=&cnt_num_vars_l.  nlobs_l=&nlobs_l.;
                   %GOTO endmac; 
         %END;  
       %ELSE  %IF &nlobs_l. ^= &nlobs_r. %THEN 
			%DO; 
   			   %LET m_return_cd=3;
			   %LET m_return_msg=Number of observation should be the same in left matrix and right vector  nlobs_l=&nlobs_l. nlobs_r=&nlobs_r.;
                   %GOTO endmac; 
         %END;
       %ELSE  %IF &cnt_num_vars_r. ^=1 %THEN 
			%DO; 
   			   %LET m_return_cd=4;
			   %LET m_return_msg=Number of numeric colums in the left matrix must be 1: cnt_num_vars_r=&cnt_num_vars_r.;
                   %GOTO endmac; 
         %END;  

	   %LET fst_var_nm_l=%SCAN(&var_lst_l. ,1, %STR( ));
	   
		/*  %PUT fst_var_nm_l=&fst_var_nm_l.;*/
 
	  /* Below we set diagonal element for "dependent" variable to some very large number. This does
       not affect results of interest which are reg. coeff (\betas) and inverse of cov matrix for \betas.
       However, large diagonal element for "dependent" variable insures that square of error (\sigma **2)
   	    calculated by PROC REG using SSCP matrix is positive.  
		 The latter prevents unnecessary warnings 
		 */

		   DATA 	__pseudo_SSCP(TYPE=SSCP RENAME=(&fst_var_nm_l.=Intercept) );
			  LENGTH _TYPE_ $ 8  _NAME_$ 32;
			   MERGE &tbl_name_left_in.(KEEP=&var_lst_l.)
							 &tbl_name_right_in.(KEEP=&var_lst_r.) END=last
					 ;
			   ARRAY vars{*} &var_lst_l. &var_lst_r.;
			   ARRAY temp{&nlobs_l.} _TEMPORARY_;

			   LENGTH var_lst_l_2 $ 32000;

			    _TYPE_="SSCP";
			    _NAME_=VNAME(vars{_N_});
				  temp{_N_}=&var_lst_r.;

				  IF  STRIP(_NAME_) =STRIP( "&fst_var_nm_l.")  THEN   _NAME_="Intercept";

						OUTPUT;

			 IF last THEN
			  	DO;
					_NAME_=VNAME(vars{_N_ +1});
					     DO i=1 TO &nlobs_l.;
						   	vars{i}=temp{i}	;
						IF I >1 THEN var_lst_l_2=CATX(" ",var_lst_l_2,VNAME(vars{i}) );
						 END;
							 vars{i} =1E16;
						     OUTPUT;

							  _TYPE_="N";
							 _NAME_=" ";
					     DO i=1 TO DIM(vars);					
						   	vars{i}=100000	;
						 END;
						     OUTPUT;

				END;/* End of IF last do-group*/

			   CALL SYMPUT("var_lst_l_2", STRIP(var_lst_l_2));
				KEEP _NAME_ _TYPE_  &var_lst_l. &var_lst_r.;
			RUN;

		/*	%PUT var_lst_l_2=&var_lst_l_2.;	*/

				ODS SELECT NONE;
	
	     ods output ParameterEstimates=__PE;
		 ods output InvXPX =__GINV;   /* This is is inverse of Inv(X'X) bordered by parm estimates*/;

    PROC REG DATA=__pseudo_SSCP
						 OUTEST=__PE_T 	 
							;
		_XBETA_:	MODEL	 &var_lst_r.=&var_lst_l_2. /  I  ;
	RUN;
	QUIT;
		   ODS SELECT ALL;
 
   DATA &tbl_name_out.;
	  SET __PE(KEEP=Variable Estimate StdErr);
 	   IF Variable="Intercept"  THEN Variable="&fst_var_nm_l.";
	  RENAME  Estimate=Value;
	   IF StdErr=. THEN 
	                          DO;
		call symput("m_return_cd", "100"); 
	   CALL SYMPUT("m_return_msg","Input Matrix is Singular. Solution is obtained using generalized Inverse and is not unique");
							 END;
	 DROP StdErr;
   RUN;

    DATA &tbl_name_out._T;
	  SET __PE_T;

	 KEEP Intercept &var_lst_l_2.;
	 RENAME Intercept=&fst_var_nm_l.;
   RUN;
   
   DATA &tbl_name_ginv_out.;
	  SET __GINV;
	    IF Variable="Intercept"  THEN Variable="&fst_var_nm_l.";
		IF 	LOWCASE(Variable) NE LOWCASE("&var_lst_r.")	;

        KEEP Variable Intercept &var_lst_l_2.;
		 RENAME Intercept=&fst_var_nm_l.;
   RUN;

   	%IF %upcase(&print.)=Y %THEN
        %DO;
			PROC PRINT DATA=&tbl_name_out. NOOBS;
			 FORMAT _NUMERIC_ D11.3;
			RUN;
		%END;

 %endmac:
                    %PUT m_return_cd=&m_return_cd. ;
                      %IF &m_return_cd. NE 0 %THEN 
					                   %DO;
						%let _tmp_err=ERROR: ;
						%IF %EVAL(&m_return_cd. - 100) =0 %THEN %let _tmp_err=Warning: ;
							  %PUT &_tmp_err. &m_return_msg. ;
									   %END;

	 %IF &DEBUG_FLAG2. NE  Y %THEN 
					                   %DO;
	PROC DATASETS NOLIST NOWARN;
	 DELETE _CONT_L _CONT_R __pseudo_SSCP __PE __PE_T __GINV __CollinDiag;
	RUN;
	QUIT;
									   %END;
   %PUT;
   %PUT -----   End of the macro solve_linear_equations_reg. ;
   %PUT;

   	OPTIONS &_DEBUG_OPTIONS_L.;
 %MEND;